// g++ tutorial.cpp -o tutorial -lsodium
#include <iostream>
#include <map>
#include <bitset>
#include <array>
#include <cmath>
#include <algorithm>
#include "sspn.h"

using namespace std;

// 遍历4位变量x，循环体为body
#define FOR_4(x, body) for (BIT4 x = 0x00; x <= 0x0f; x++) body

template<int b>
int CountOne(uint64_t data) {
    int one = 0;
    for (int i = 0; i < b; i++) {
        if ((data >> i) & 0x1 == 0x1) {
            one++;
        }
    }
    return one;
}

int main() {
    printf("`c:Start.`");
    // 生成密钥
    BIT16 rk[5];
    subkey(rk);
    // 生成10000组明密文对
    #define PAIR_NUM 100000
    BIT16 plains[PAIR_NUM], ciphers[PAIR_NUM];
    for (int i = 0; i < PAIR_NUM; i++) {
        plains[i] = random16();
        ciphers[i] = encrypt(plains[i], rk);
    }
    // 保存结果
    struct Result {
        double bias;
        BIT4 k2;
        BIT4 k4;
    };
    array<Result, 256> results;
    int index = 0;

    // 枚举猜测的密钥值
    FOR_4(k2, {
        FOR_4(k4, {
            int count = 0;
            for (int i = 0; i < PAIR_NUM; i++) {
                // 计算出最后一轮s盒的输出
                BIT4 c1;
                BIT4 c2;
                BIT4 c3;
                BIT4 c4;
                split(ciphers[i], &c1, &c2, &c3, &c4);
                BIT4 so2 = c2 ^ k2;
                BIT4 so4 = c4 ^ k4;
                // 计算最后一轮s盒的输入，即U4
                BIT4 u2 = sbox_inverse(so2);
                BIT4 u4 = sbox_inverse(so4);
                // 代入线性表达式，检查线性表达式是否成立
                BIT16 plain_mask = 0x0b00;
                BIT4 u_mask = 0x5;
                int value = (CountOne<16>(plains[i] & plain_mask) + CountOne<4>(u2 & u_mask) + CountOne<4>(u4 & u_mask)) % 2;
                if (value == 0) {
                    count++;
                }
            }
            results[index].bias = abs((double)count / (double)PAIR_NUM - 0.5);
            results[index].k2 = k2;
            results[index].k4 = k4;
            index++;
        })
    })
    // 找出bias最接近1/32的猜测
    sort(results.begin(), results.end(), [](Result a, Result b) {
        return abs(a.bias - 1.0f / 32.0f) < abs(b.bias - 1.0f / 32.0f);
    });
    // 打印正确的key
    printf("last round key: %04x, 1/32=%3f\n", rk[4], 1.0f / 32.0f);
    // 找出前十名
    for (int i = 0; i < 10; i++) {
        printf("bias = %.3f, k2 = %04x, k4 = %04x\n", results[i].bias, results[i].k2, results[i].k4);
    }
    printf("`p:100:100`");
    printf("`c:Done.`");
}